(in-package :user)

(defclass appointment ()
  ((appointment-time :initarg :time :accessor appointment-time)
   (minutes-long :initarg :minutes-long :accessor minutes-long)
   (doctor :initarg :doctor :accessor doctor)))

(defmethod print-object ((a appointment) stream)
  (if *print-escape*
      (print-unreadable-object (a stream :identity t :type t)
        (format stream "~A" (appointment-time a)))
    (format stream "~A" (appointment-time a))))

(defclass patient-appointment (appointment)
  ((patient :initarg :patient :accessor patient)))

(defclass meeting (appointment)
  ())

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; For now, the schedule is a list.
(defvar *schedule* nil)

(defmethod conflict-p ((a1 appointment) (a2 appointment))
  ;; Doctors can be double-booked.
  nil)

(defmethod conflict-p ((a1 patient-appointment) (a2 patient-appointment))
  ;; Patients cannot be double-booked.
  (cond ((not (eql (patient a1) (patient a2)))
         nil)
        ((date-equalp (appointment-time a1) (appointment-time a2))
         ;; Both start at the same time.
         t)
        ((and (date-lessp (appointment-time a1) (appointment-time a2))
              (date-lessp (appointment-time a2) 
                          (date-add (appointment-time a1) (minutes-long a1))))
         ;; A1 starts before A2 starts and ends after A2 starts.
         t)
        ((and (date-lessp (appointment-time a2) (appointment-time a1))
              (date-lessp (appointment-time a1)
                          (date-add (appointment-time a2) (minutes-long a2))))
         ;; A2 starts before A1 starts and ends after A1 starts.
         t)))

(defmethod can-schedule ((a appointment))
  (dolist (item *schedule*)
    (if (conflict-p item a) (return-from can-schedule nil)))
  t)    

(defmethod schedule ((a appointment))
  (when (can-schedule a) 
    (push a *schedule*) 
    a))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod cancel ((a appointment))
  ;; use destructive 'delete' instead of 'remove' (no need to copy list)
  (setq *schedule* (delete a *schedule*))
  a)

(defmethod reschedule ((a appointment) new-time)
  (let ((old-time (appointment-time a)))
    (setf (appointment-time a) new-time)
    (cond ((can-schedule a) 
           (push a *schedule*) 
           a)
          (t
           (setf (appointment-time a) old-time)
           nil))))

(defmethod patient-appointment-p ((appt t)) nil)
(defmethod patient-appointment-p ((appt patient-appointment)) t)

(defun patient-appointments (patient)
  (let ((result nil))
    (dolist (item *schedule*)
      (when (patient-appointment-p item)
        (when (eql (patient item) patient)
          (push item result))))
    (sort result #'(lambda (a1 a2) 
                     (date-lessp (appointment-time a1) 
                                 (appointment-time a2))))))

(defun doctor-appointments (doctor)
  (let ((result nil))
    (dolist (item *schedule*)
      (when (eql (doctor item) doctor)
        (push item result)))
    (sort result #'(lambda (a1 a2) 
                     (date-lessp (appointment-time a1) 
                                 (appointment-time a2))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Test data

(defparameter a1 (make-instance 'patient-appointment 
                   :doctor 'welby 
                   :patient 'sally
                   :time (make-date +september+ 4 2003 8 0)  ; 8am with Welby
                   :minutes-long 30))

(defparameter a2 (make-instance 'patient-appointment 
                   :doctor 'dre 
                   :patient 'sally
                   :time (make-date +september+ 4 2003 9 0)  ; 9am with Dre
                   :minutes-long 30))

(defparameter a3 (make-instance 'patient-appointment 
                   :doctor 'welby 
                   :patient 'bob
                   :time (make-date +september+ 4 2003 8 0)  ; 8am with Welby
                   :minutes-long 30))

(defparameter a4 (make-instance 'patient-appointment 
                   :doctor 'dre 
                   :patient 'sally
                   :time (make-date +september+ 4 2003 8 0)  ; 8am with Dre
                   :minutes-long 30))

(schedule a1)
(schedule a2)
(schedule a3)
(schedule a4) ; conflicting appointment

(patient-appointments 'sally)
(patient-appointments 'bob)
(doctor-appointments 'welby)
(doctor-appointments 'dre)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Some ways of saving the schedule:
;;; 1. Tab or comma delimited ascii file
;;; 2. Print Lisp code to a file that, when loaded, recreates the schedule.
;;; 3. Use Allegro fasl-read and fasl-write to generate binary files
;;; 4. Use a relational database (MS Access, Oracle, SQL Server, MySQL) 
;;;    and build an ODBC interface to it
;;; 5. Use an object-oriented database (AllegroStore)
;;; 6. Save your entire lisp image, data and all

;;; A relational database is the option that will run the fastest
;;; and scale to very large data sets.  Drawback is that you have
;;; to cast your data as tables with a fixed number of columns.
;;; For example, how would you store APPOINTMENT and PATIENT-APPOINTMENT
;;; in the same table, since one has more slots than the other?

;;; An object-oriented database is slower but extremely general.

;;; Tab or comma delimited ascii files are inexpensive (free!) but
;;; you have the same limitations as a relational database, and they
;;; don't work well with lots of data.

;;; Fasl-read and Fasl-write nicely handle lisp objects (even circular
;;; data structures).  Not portable to other languages, other operating
;;; systems, or even other versions of Allegro.

